@@ -10,6 +10,7 @@ import re
 import requests
 from os.path import expanduser
 import copy
+import urllib
 
 __author__ = 'Reinoud van Leeuwen'
 __copyright__ = "Copyright 2013, Reinoud van Leeuwen"
@@ -23,7 +24,7 @@ CONVERTABLEFIELDS = {
                  'tax_shifted': 'bool',
                  'lastinvoice': 'date',
                  'top': 'int',
-                 'stddiscount': 'int',
+                 'stddiscount': 'float',
                  'notes_on_invoice': 'bool',
                  'active': 'bool',
                  'default_email': 'int',
@@ -41,7 +42,7 @@ CONVERTABLEFIELDS = {
                  'uncollectible': 'date',
                  'lastreminder': 'date',
                  'open': 'float',
-                 'paiddate': 'float',
+                 'paiddate': 'date',
                  'duedate': 'date',
                  'overwrite_if_exist': 'bool',
                  'initialdate': 'date',
@@ -130,6 +131,11 @@ class FactuursturenConversionError(FactuursturenError):
 class FactuursturenWrongCall(FactuursturenError):
     pass
 
+class FactuursturenNotFound(FactuursturenError):
+    pass
+
+class FactuursturenNoMoreApiCalls(FactuursturenError):
+    pass
 
 class Client:
     """
@@ -144,19 +150,37 @@ class Client:
                  protocol='https',
                  apipath='/api',
                  version='v0'):
+        """
+        initialize object
+
+        When apikey and username are not present, look for INI-style file .factuursturen_rc
+        in current directory and homedirectory to find those values there.
+        when only username is present, try to find apikey in configfilesection where it is defined
+
+        :param apikey: APIkey (string) as generated online on the website http://www.factuursturen.nl
+        :param username: accountname for the website
+        :param configsection: section in file ~/.factuursturen_rc where apikey and username should be present
+        """
         self._url = protocol + '://' + host + apipath + '/' + version + '/'
 
         # try to read auth details from file when not passed
+        config = ConfigParser.RawConfigParser()
+        config.read(['.factuursturen_rc', expanduser('~/.factuursturen_rc')])
         if (not apikey) and (not username):
             try:
-                config = ConfigParser.RawConfigParser()
-                config.read(['.factuursturen_rc', expanduser('~/.factuursturen_rc')])
                 self._apikey = config.get(configsection, 'apikey')
                 self._username = config.get(configsection, 'username')
             except ConfigParser.NoSectionError:
                 raise FactuursturenNoAuth ('key and username not given, nor found in .factuursturen_rc or ~/.factuursturen_rc')
             except ConfigParser.NoOptionError:
                 raise FactuursturenNoAuth ('no complete auth found')
+        elif username and (not apikey):
+            self._username = username
+            for section in config.sections():
+                if config.get(section, 'username') == username:
+                    self._apikey = config.get(section, 'apikey')
+            if not self._apikey:
+                raise FactuursturenNoAuth ('no apikey found for username {}'.format(username))
         else:
             if not (apikey and username):
                 raise FactuursturenNoAuth ('no complete auth passed to factuursturen.Client')
@@ -198,6 +222,8 @@ class Client:
             raise FactuursturenConversionError('cannot convert {} to float'.format(string))
 
     def _string2date(self, string):
+        if string == '':
+            return None
         try:
             return datetime.strptime(string, '%Y-%m-%d')
         except ValueError:
@@ -225,9 +251,13 @@ class Client:
 
     def _convertstringfields_in_dict(self, adict, function, direction):
         """convert fields of a single dict either from or to strings
-        :param adict:
-        :param function:
-        :param direction:
+
+        fieldnames to convert are read from CONVERTIBLEFIELDS dict, which
+        is in essence a datadictionary for this API
+
+        :param adict: dictionary to convert
+        :param function: callable function in the API ('clients', 'products' etc)
+        :param direction: either 'tostring' or 'fromstring'
         """
         if direction not in self._convertfunctions:
             raise FactuursturenWrongCall ('_convertstringfields_in_dict called with {}'.format(direction))
@@ -238,15 +268,21 @@ class Client:
                     # on conversion direction, this is the source or the target
                     target = CONVERTABLEFIELDS[function][key]
                     conversion_function = self._convertfunctions[direction][target]
-                    adict[key] = conversion_function(value)
+                    try:
+                        adict[key] = conversion_function(value)
+                    except FactuursturenConversionError:
+                        print "key = {}, value = {}, direction = {}, target = {}".format(key, value, direction, target)
+                        raise BaseException
         return adict
 
     def _convertstringfields_in_list_of_dicts(self, alist, function, direction):
         """convert each dict in the list
 
-        :param alist:
-        :param function:
-        :param direction:
+        Basically, a loop over the function _convertstringfields_in_dict
+
+        :param alist: a list of dicts
+        :param function: callable function in the API ('clients', 'products' etc)
+        :param direction: either 'tostring' or 'fromstring'
         """
         if direction not in self._convertfunctions:
             raise FactuursturenWrongCall ('_convertstringfields_in_list_of_dicts called with {}'.format(direction))
@@ -257,6 +293,8 @@ class Client:
     def _flatten(self, adict, parent_key=''):
         """flatten a nested dict
 
+        The API expects nested dicts to be flattened when posting
+
         {'lines': {'line1': {'amount': 1,
                              'tax': 21},
                    'line2': {'amount': 2,
@@ -270,8 +308,8 @@ class Client:
          'lines[line2][tax]': 21
         }
 
-        :param adict:
-        :param parent_key:
+        :param adict: a nested dict
+        :param parent_key: should be empty, used for recursion
         """
         items = []
         for k, v in adict.items():
@@ -289,7 +327,7 @@ class Client:
         with 'lines[0][amount_desc]'
         (keeping the same value)
 
-        :param adict:
+        :param adict: dictionary to be changed
         """
         for key, val in adict.items():
             fields = re.split('\]\[', key)
@@ -304,14 +342,22 @@ class Client:
     def _prepare_for_send(self, adict, function):
         """fix dict so it can be posted
 
-        :param adict:
-        :param function:
+        :param adict: dictionary to be posted
+        :param function: callable function from the API ('clients', 'products', etc)
         """
         adict = self._convertstringfields_in_dict(adict, function, 'tostring')
         adict = self._flatten(adict)
         adict = self._fixkeynames(adict)
         return adict
 
+    def _escape_characters(self, string):
+        """escape unsafe webcharacters to use in API call
+
+        by default urllib considers '/' as safe, override the default for the second argument by
+        considering nothing safe
+        """
+        return urllib.quote(str(string), safe='')
+
     @property
     def remaining(self):
         """return remaining allowed API calls (for this hour)"""
@@ -326,8 +372,9 @@ class Client:
         """Generic wrapper for all POSTable functions
 
         errors from server during post (like wrong values) are propagated to the exceptionclass
-        :param function:
-        :param objData_local:
+
+        :param function: callabe function from the API ('clients', 'products', etc)
+        :param objData: data to be posted
         """
         fullUrl = self._url + function
         objData_local = copy.deepcopy(objData)
@@ -352,11 +399,12 @@ class Client:
         """Generic wrapper for all PUTable functions
 
         errors from server during post (like wrong values) are propagated to the exceptionclass
-        :param function:
-        :param objId:
-        :param objData:
+
+        :param function: callabe function from the API ('clients', 'products', etc)
+        :param objId: id of object to be put (usually retrieved from the API)
+        :param objData: data to be posted. All required fields should be present, or the API will not accept the changes
         """
-        fullUrl = self._url + function + '/{objId}'.format(objId=objId)
+        fullUrl = self._url + function + '/{objId}'.format(objId=self._escape_characters(objId))
 
         if function not in API['putters']:
             raise FactuursturenPostError("{function} not in available PUTable functions".format(function=function))
@@ -379,10 +427,10 @@ class Client:
         """Generic wrapper for all DELETEable functions
 
         errors from server during post (like wrong values) are propagated to the exceptionclass
-        :param function:
-        :param objId:
+        :param function: callabe function from the API ('clients', 'products', etc)
+        :param objId: id of object to be put (usually retrieved from the API)
         """
-        fullUrl = self._url + function + '/{objId}'.format(objId=objId)
+        fullUrl = self._url + function + '/{objId}'.format(objId=self._escape_characters(objId))
 
         if function not in API['deleters']:
             raise FactuursturenPostError("{function} not in available DELETEable functions".format(function=function))
@@ -400,8 +448,11 @@ class Client:
     def get(self, function, objId=None):
         """Generic wrapper for all GETtable functions
 
-        :param function:
-        :param objId:
+        when no objId is passed, retrieve all objects (in a list of dicts)
+        when objId is passed, only retrieve a single object (in a single dict)
+
+        :param function: callabe function from the API ('clients', 'products', etc)
+        :param objId: id of object to be put (usually retrieved from the API)
         """
 
         # TODO: some errorchecking:
@@ -417,7 +468,7 @@ class Client:
             raise FactuursturenGetError("{function} not in available GETtable functions".format(function=function))
 
         if objId:
-            fullUrl += '/{objId}'.format(objId=objId)
+            fullUrl += '/{objId}'.format(objId=self._escape_characters(objId))
 
         response = requests.get(fullUrl,
                                 auth=(self._username, self._apikey),
@@ -429,18 +480,26 @@ class Client:
         # when the functioncall was 'clients/<id>
 
         singlefunction = function[:-1]
+        self._remaining = int(response.headers['x-ratelimit-remaining'])
+
         if response.ok:
-            self._remaining = int(response.headers['x-ratelimit-remaining'])
+            if function == 'invoices_pdf':
+                return response.content
             try:
                 raw_structure = response.json()
-                if isinstance(raw_structure, dict):
-
-                    retval = self._convertstringfields_in_dict(raw_structure[singlefunction], function, 'fromstring')
-                else:
+                if objId is None:
                     retval = self._convertstringfields_in_list_of_dicts(raw_structure, function, 'fromstring')
-            except FactuursturenError:
+                else:
+                    retval = self._convertstringfields_in_dict(raw_structure[singlefunction], function, 'fromstring')
+            except FactuursturenError as error:
+                print error
                 retval = response.content
             return retval
         else:
             # TODO: more checking
-            raise FactuursturenEmptyResult (response.content)
+            if response.status_code == 404:
+                raise FactuursturenNotFound (response.content)
+            elif self._remaining == 0:
+                raise FactuursturenNoMoreApiCalls ('limit of API calls reached.')
+            else:
+                raise FactuursturenEmptyResult (response.content)
