@@ -237,7 +237,10 @@ class test_client(TestCase):
                         u'name': u'Test produkt via API',
                         u'price': 123.450,
                         u'taxes': 21}
-        added_id = int(fact.post('products', test_product))
+        try:
+            added_id = int(fact.post('products', test_product))
+        except factuursturen.FactuursturenWrongPostvalue:
+            pytest.skip("skipping online tests, wrong authentication in  ~/.factuursturen_rc")
         self.assertTrue(fact.ok)
 
         try:
@@ -262,6 +265,8 @@ class test_client(TestCase):
         try:
             test_returned_product = fact.get('products', added_id)
             self.fail('an exception should be raised when trying to get a deleted object')
-        except factuursturen.FactuursturenEmptyResult:
+        except factuursturen.FactuursturenNotFound:
             self.assertFalse(fact.ok)
+        except Exception as error:
+            self.fail('expecting an FactuursturenNotFound exception, got {}'.format(error))
 
